@@ -162,7 +162,7 @@ void QWaylandWindow::initWindow()
                     name.chop(8);
                 mShellSurface->setAppId(name);
             } else {
-                QFileInfo fi = QCoreApplication::instance()->applicationFilePath();
+                QFileInfo fi = QFileInfo(QCoreApplication::instance()->applicationFilePath());
                 QStringList domainName =
                         QCoreApplication::instance()->organizationDomain().split(QLatin1Char('.'),
                                                                                  Qt::SkipEmptyParts);
@@ -188,15 +188,13 @@ void QWaylandWindow::initWindow()
     // Enable high-dpi rendering. Scale() returns the screen scale factor and will
     // typically be integer 1 (normal-dpi) or 2 (high-dpi). Call set_buffer_scale()
     // to inform the compositor that high-resolution buffers will be provided.
-    if (mDisplay->compositorVersion() >= 3)
-        mSurface->set_buffer_scale(scale());
+    if (mSurface->version() >= 3)
+        mSurface->set_buffer_scale(mScale);
 
-    if (QScreen *s = window()->screen())
-        setOrientationMask(s->orientationUpdateMask());
     setWindowFlags(window()->flags());
     QRect geometry = windowGeometry();
     if (geometry.isEmpty())
-        setGeometry_helper(QRect(QPoint(), QSize(500,500)));
+        setGeometry_helper(defaultGeometry());
     else
         setGeometry_helper(geometry);
     setMask(window()->mask());
@@ -263,6 +261,7 @@ void QWaylandWindow::reset()
     mFrameCallbackElapsedTimer.invalidate();
     mWaitingForFrameCallback = false;
     mFrameCallbackTimedOut = false;
+    mWaitingToApplyConfigure = false;
 
     mMask = QRegion();
     mQueuedBuffer = nullptr;
@@ -311,7 +310,7 @@ void QWaylandWindow::setWindowTitle(const QString &title)
         // three bytes when converted to utf-8 (which is what libwayland uses), so divide by three.
         const int maxLength = libwaylandMaxBufferSize / 3 - 100;
 
-        auto truncated = QStringRef(&formatted).left(maxLength);
+        auto truncated = QStringView{formatted}.left(maxLength);
         if (truncated.length() < formatted.length()) {
             qCWarning(lcQpaWayland) << "Window titles longer than" << maxLength << "characters are not supported."
                                     << "Truncating window title (from" << formatted.length() << "chars)";
@@ -319,7 +318,7 @@ void QWaylandWindow::setWindowTitle(const QString &title)
         mShellSurface->setTitle(truncated.toString());
     }
 
-    if (mWindowDecoration && window()->isVisible())
+    if (mWindowDecorationEnabled && window()->isVisible())
         mWindowDecoration->update();
 }
 
@@ -327,10 +326,15 @@ void QWaylandWindow::setWindowIcon(const QIcon &icon)
 {
     mWindowIcon = icon;
 
-    if (mWindowDecoration && window()->isVisible())
+    if (mWindowDecorationEnabled && window()->isVisible())
         mWindowDecoration->update();
 }
 
+QRect QWaylandWindow::defaultGeometry() const
+{
+    return QRect(QPoint(), QSize(500,500));
+}
+
 void QWaylandWindow::setGeometry_helper(const QRect &rect)
 {
     QSize minimum = windowMinimumSize();
@@ -356,7 +360,7 @@ void QWaylandWindow::setGeometry(const QRect &rect)
     setGeometry_helper(rect);
 
     if (window()->isVisible() && rect.isValid()) {
-        if (mWindowDecoration)
+        if (mWindowDecorationEnabled)
             mWindowDecoration->update();
 
         if (mResizeAfterSwap && windowType() == Egl && mSentInitialResize)
@@ -380,6 +384,16 @@ void QWaylandWindow::setGeometry(const QRect &rect)
 void QWaylandWindow::resizeFromApplyConfigure(const QSize &sizeWithMargins, const QPoint &offset)
 {
     QMargins margins = frameMargins();
+
+    // Exclude shadows from margins once they are excluded from window geometry
+    // 1) First resizeFromApplyConfigure() call will have sizeWithMargins equal to surfaceSize()
+    //    which has full margins (shadows included).
+    // 2) Following resizeFromApplyConfigure() calls should have sizeWithMargins equal to
+    //    windowContentGeometry() which excludes shadows, therefore in this case we have to
+    //    exclude them too in order not to accidentally apply smaller size to the window.
+    if (mWindowDecorationEnabled && (sizeWithMargins != surfaceSize()))
+        margins = mWindowDecoration->margins(QWaylandAbstractDecoration::ShadowsExcluded);
+
     int widthWithoutMargins = qMax(sizeWithMargins.width() - (margins.left() + margins.right()), 1);
     int heightWithoutMargins = qMax(sizeWithMargins.height() - (margins.top() + margins.bottom()), 1);
     QRect geometry(windowGeometry().topLeft(), QSize(widthWithoutMargins, heightWithoutMargins));
@@ -400,7 +414,7 @@ void QWaylandWindow::sendExposeEvent(const QRect &rect)
 }
 
 
-static QVector<QPointer<QWaylandWindow>> activePopups;
+static QList<QPointer<QWaylandWindow>> activePopups;
 
 void QWaylandWindow::closePopups(QWaylandWindow *parent)
 {
@@ -556,12 +570,12 @@ void QWaylandWindow::sendRecursiveExposeEvent()
 
 void QWaylandWindow::attach(QWaylandBuffer *buffer, int x, int y)
 {
-    Q_ASSERT(!buffer->committed());
     QReadLocker locker(&mSurfaceLock);
     if (mSurface == nullptr)
         return;
 
     if (buffer) {
+        Q_ASSERT(!buffer->committed());
         handleUpdate();
         buffer->setBusy();
 
@@ -583,7 +597,11 @@ void QWaylandWindow::damage(const QRect &rect)
     if (mSurface == nullptr)
         return;
 
-    mSurface->damage(rect.x(), rect.y(), rect.width(), rect.height());
+    const qreal s = scale();
+    if (mSurface->version() >= 4)
+        mSurface->damage_buffer(qFloor(s * rect.x()), qFloor(s * rect.y()), qCeil(s * rect.width()), qCeil(s * rect.height()));
+    else
+        mSurface->damage(rect.x(), rect.y(), rect.width(), rect.height());
 }
 
 void QWaylandWindow::safeCommit(QWaylandBuffer *buffer, const QRegion &damage)
@@ -619,8 +637,14 @@ void QWaylandWindow::commit(QWaylandBuffer *buffer, const QRegion &damage)
         return;
 
     attachOffset(buffer);
-    for (const QRect &rect: damage)
-        mSurface->damage(rect.x(), rect.y(), rect.width(), rect.height());
+    if (mSurface->version() >= 4) {
+        const qreal s = scale();
+        for (const QRect &rect: damage)
+            mSurface->damage_buffer(qFloor(s * rect.x()), qFloor(s * rect.y()), qCeil(s * rect.width()), qCeil(s * rect.height()));
+    } else {
+        for (const QRect &rect: damage)
+            mSurface->damage(rect.x(), rect.y(), rect.width(), rect.height());
+    }
     Q_ASSERT(!buffer->committed());
     buffer->setCommitted();
     mSurface->commit();
@@ -649,24 +673,25 @@ void QWaylandWindow::handleFrameCallback()
 
     // The rest can wait until we can run it on the correct thread
     if (!mWaitingForUpdateDelivery) {
-        auto doHandleExpose = [this]() {
-            bool wasExposed = isExposed();
-            mFrameCallbackTimedOut = false;
-            if (!wasExposed && isExposed()) // Did setting mFrameCallbackTimedOut make the window exposed?
-                sendExposeEvent(QRect(QPoint(), geometry().size()));
-            if (wasExposed && hasPendingUpdateRequest())
-                deliverUpdateRequest();
-
-            mWaitingForUpdateDelivery = false;
-        };
-
         // Queued connection, to make sure we don't call handleUpdate() from inside waitForFrameSync()
         // in the single-threaded case.
         mWaitingForUpdateDelivery = true;
-        QMetaObject::invokeMethod(this, doHandleExpose, Qt::QueuedConnection);
+        QMetaObject::invokeMethod(this, &QWaylandWindow::doHandleFrameCallback, Qt::QueuedConnection);
     }
 }
 
+void QWaylandWindow::doHandleFrameCallback()
+{
+    bool wasExposed = isExposed();
+    mFrameCallbackTimedOut = false;
+    if (!wasExposed && isExposed()) // Did setting mFrameCallbackTimedOut make the window exposed?
+        sendExposeEvent(QRect(QPoint(), geometry().size()));
+    if (wasExposed && hasPendingUpdateRequest())
+        deliverUpdateRequest();
+
+    mWaitingForUpdateDelivery = false;
+}
+
 bool QWaylandWindow::waitForFrameSync(int timeout)
 {
     QMutexLocker locker(mFrameQueue.mutex);
@@ -684,7 +709,7 @@ bool QWaylandWindow::waitForFrameSync(int timeout)
 
 QMargins QWaylandWindow::frameMargins() const
 {
-    if (mWindowDecoration)
+    if (mWindowDecorationEnabled)
         return mWindowDecoration->margins();
     return QPlatformWindow::frameMargins();
 }
@@ -703,7 +728,12 @@ QSize QWaylandWindow::surfaceSize() const
  */
 QRect QWaylandWindow::windowContentGeometry() const
 {
-    return QRect(QPoint(), surfaceSize());
+    QMargins shadowMargins;
+
+    if (mWindowDecorationEnabled)
+        shadowMargins = mWindowDecoration->margins(QWaylandAbstractDecoration::ShadowsOnly);
+
+    return QRect(QPoint(shadowMargins.left(), shadowMargins.top()), surfaceSize().shrunkBy(shadowMargins));
 }
 
 /*!
@@ -747,7 +777,7 @@ QWaylandScreen *QWaylandWindow::waylandScreen() const
 void QWaylandWindow::handleContentOrientationChange(Qt::ScreenOrientation orientation)
 {
     QReadLocker locker(&mSurfaceLock);
-    if (mDisplay->compositorVersion() < 2 || mSurface == nullptr)
+    if (mSurface == nullptr || mSurface->version() < 2)
         return;
 
     wl_output_transform transform;
@@ -824,9 +854,14 @@ bool QWaylandWindow::createDecoration()
     if (!mShellSurface || !mShellSurface->wantsDecorations())
         decoration = false;
 
-    bool hadDecoration = mWindowDecoration;
+    bool hadDecoration = mWindowDecorationEnabled;
     if (decoration && !decorationPluginFailed) {
-        if (!mWindowDecoration) {
+        if (!mWindowDecorationEnabled) {
+            if (mWindowDecoration) {
+                delete mWindowDecoration;
+                mWindowDecoration = nullptr;
+            }
+
             QStringList decorations = QWaylandDecorationFactory::keys();
             if (decorations.empty()) {
                 qWarning() << "No decoration plugins available. Running with no decorations.";
@@ -855,13 +890,13 @@ bool QWaylandWindow::createDecoration()
                 return false;
             }
             mWindowDecoration->setWaylandWindow(this);
+            mWindowDecorationEnabled = true;
         }
     } else {
-        delete mWindowDecoration;
-        mWindowDecoration = nullptr;
+        mWindowDecorationEnabled = false;
     }
 
-    if (hadDecoration != (bool)mWindowDecoration) {
+    if (hadDecoration != mWindowDecorationEnabled) {
         for (QWaylandSubSurface *subsurf : qAsConst(mChildren)) {
             QPoint pos = subsurf->window()->geometry().topLeft();
             QMargins m = frameMargins();
@@ -882,7 +917,7 @@ bool QWaylandWindow::createDecoration()
 
 QWaylandAbstractDecoration *QWaylandWindow::decoration() const
 {
-    return mWindowDecoration;
+    return mWindowDecorationEnabled ? mWindowDecoration : nullptr;
 }
 
 static QWaylandWindow *closestShellSurfaceWindow(QWindow *window)
@@ -913,7 +948,7 @@ QWaylandWindow *QWaylandWindow::transientParent() const
 void QWaylandWindow::handleMouse(QWaylandInputDevice *inputDevice, const QWaylandPointerEvent &e)
 {
     if (e.type == QEvent::Leave) {
-        if (mWindowDecoration) {
+        if (mWindowDecorationEnabled) {
             if (mMouseEventsInContentArea)
                 QWindowSystemInterface::handleLeaveEvent(window());
         } else {
@@ -925,7 +960,7 @@ void QWaylandWindow::handleMouse(QWaylandInputDevice *inputDevice, const QWaylan
         return;
     }
 
-    if (mWindowDecoration) {
+    if (mWindowDecorationEnabled) {
         handleMouseEventWithDecoration(inputDevice, e);
     } else {
         switch (e.type) {
@@ -949,16 +984,143 @@ void QWaylandWindow::handleMouse(QWaylandInputDevice *inputDevice, const QWaylan
 
 #if QT_CONFIG(cursor)
     if (e.type == QEvent::Enter) {
-        QRect contentGeometry = windowContentGeometry().marginsRemoved(frameMargins());
+        QRect contentGeometry = QRect(QPoint(), surfaceSize()).marginsRemoved(frameMargins());
         if (contentGeometry.contains(e.local.toPoint()))
             restoreMouseCursor(inputDevice);
     }
 #endif
 }
 
-bool QWaylandWindow::touchDragDecoration(QWaylandInputDevice *inputDevice, const QPointF &local, const QPointF &global, Qt::TouchPointState state, Qt::KeyboardModifiers mods)
+#ifndef QT_NO_GESTURES
+void QWaylandWindow::handleSwipeGesture(QWaylandInputDevice *inputDevice,
+                                        const QWaylandPointerGestureSwipeEvent &e)
 {
-    if (!mWindowDecoration)
+    switch (e.state) {
+        case Qt::GestureStarted:
+            if (mGestureState != GestureNotActive)
+                qCWarning(lcQpaWaylandInput) << "Unexpected gesture state. Gestures will act weird.";
+
+            if (mWindowDecorationEnabled && !mMouseEventsInContentArea) {
+                // whole gesture sequence will be ignored
+                mGestureState = GestureActiveInDecoration;
+                return;
+            }
+
+            mGestureState = GestureActiveInContentArea;
+            QWindowSystemInterface::handleGestureEvent(window(), e.timestamp,
+                                                       inputDevice->mTouchPadDevice,
+                                                       Qt::BeginNativeGesture,
+                                                       e.local, e.global, e.fingers);
+            break;
+        case Qt::GestureUpdated:
+            if (mGestureState != GestureActiveInContentArea)
+                return;
+
+            if (!e.delta.isNull()) {
+                QWindowSystemInterface::handleGestureEventWithValueAndDelta(
+                            window(), e.timestamp, inputDevice->mTouchPadDevice,
+                            Qt::PanNativeGesture,
+                            0, e.delta, e.local, e.global, e.fingers);
+            }
+            break;
+        case Qt::GestureFinished:
+        case Qt::GestureCanceled:
+            if (mGestureState == GestureActiveInDecoration) {
+                mGestureState = GestureNotActive;
+                return;
+            }
+
+            if (mGestureState != GestureActiveInContentArea)
+                qCWarning(lcQpaWaylandInput) << "Unexpected gesture state. Gestures will act weird.";
+
+            mGestureState = GestureNotActive;
+
+            // There's currently no way to expose cancelled gestures to the rest of Qt, so
+            // this part of information is lost.
+            QWindowSystemInterface::handleGestureEvent(window(), e.timestamp,
+                                                       inputDevice->mTouchPadDevice,
+                                                       Qt::EndNativeGesture,
+                                                       e.local, e.global, e.fingers);
+            break;
+        default:
+            break;
+    }
+}
+
+void QWaylandWindow::handlePinchGesture(QWaylandInputDevice *inputDevice,
+                                        const QWaylandPointerGesturePinchEvent &e)
+{
+    switch (e.state) {
+        case Qt::GestureStarted:
+            if (mGestureState != GestureNotActive)
+                qCWarning(lcQpaWaylandInput) << "Unexpected gesture state. Gestures will act weird.";
+
+            if (mWindowDecorationEnabled && !mMouseEventsInContentArea) {
+                // whole gesture sequence will be ignored
+                mGestureState = GestureActiveInDecoration;
+                return;
+            }
+
+            mGestureState = GestureActiveInContentArea;
+            QWindowSystemInterface::handleGestureEvent(window(), e.timestamp,
+                                                       inputDevice->mTouchPadDevice,
+                                                       Qt::BeginNativeGesture,
+                                                       e.local, e.global, e.fingers);
+            break;
+        case Qt::GestureUpdated:
+            if (mGestureState != GestureActiveInContentArea)
+                return;
+
+            if (!e.delta.isNull()) {
+                QWindowSystemInterface::handleGestureEventWithValueAndDelta(
+                            window(), e.timestamp, inputDevice->mTouchPadDevice,
+                            Qt::PanNativeGesture,
+                            0, e.delta, e.local, e.global, e.fingers);
+            }
+            if (e.rotation_delta != 0) {
+                QWindowSystemInterface::handleGestureEventWithRealValue(window(), e.timestamp,
+                                                                        inputDevice->mTouchPadDevice,
+                                                                        Qt::RotateNativeGesture,
+                                                                        e.rotation_delta,
+                                                                        e.local, e.global, e.fingers);
+            }
+            if (e.scale_delta != 0) {
+                QWindowSystemInterface::handleGestureEventWithRealValue(window(), e.timestamp,
+                                                                        inputDevice->mTouchPadDevice,
+                                                                        Qt::ZoomNativeGesture,
+                                                                        e.scale_delta,
+                                                                        e.local, e.global, e.fingers);
+            }
+            break;
+        case Qt::GestureFinished:
+        case Qt::GestureCanceled:
+            if (mGestureState == GestureActiveInDecoration) {
+                mGestureState = GestureNotActive;
+                return;
+            }
+
+            if (mGestureState != GestureActiveInContentArea)
+                qCWarning(lcQpaWaylandInput) << "Unexpected gesture state. Gestures will act weird.";
+
+            mGestureState = GestureNotActive;
+
+            // There's currently no way to expose cancelled gestures to the rest of Qt, so
+            // this part of information is lost.
+            QWindowSystemInterface::handleGestureEvent(window(), e.timestamp,
+                                                       inputDevice->mTouchPadDevice,
+                                                       Qt::EndNativeGesture,
+                                                       e.local, e.global, e.fingers);
+            break;
+        default:
+            break;
+    }
+}
+#endif // #ifndef QT_NO_GESTURES
+
+
+bool QWaylandWindow::touchDragDecoration(QWaylandInputDevice *inputDevice, const QPointF &local, const QPointF &global, QEventPoint::State state, Qt::KeyboardModifiers mods)
+{
+    if (!mWindowDecorationEnabled)
         return false;
     return mWindowDecoration->handleTouch(inputDevice, local, global, state, mods);
 }
@@ -1034,7 +1196,7 @@ void QWaylandWindow::handleScreensChanged()
     int scale = newScreen->isPlaceholder() ? 1 : static_cast<QWaylandScreen *>(newScreen)->scale();
     if (scale != mScale) {
         mScale = scale;
-        if (mSurface && mDisplay->compositorVersion() >= 3)
+        if (mSurface && mSurface->version() >= 3)
             mSurface->set_buffer_scale(mScale);
         ensureSize();
     }
@@ -1080,14 +1242,14 @@ bool QWaylandWindow::isActive() const
     return mDisplay->isWindowActivated(this);
 }
 
-int QWaylandWindow::scale() const
+qreal QWaylandWindow::scale() const
 {
-    return mScale;
+    return devicePixelRatio();
 }
 
 qreal QWaylandWindow::devicePixelRatio() const
 {
-    return mScale;
+    return qreal(mScale);
 }
 
 bool QWaylandWindow::setMouseGrabEnabled(bool grab)
@@ -1101,6 +1263,21 @@ bool QWaylandWindow::setMouseGrabEnabled(bool grab)
     return true;
 }
 
+QWaylandWindow::ToplevelWindowTilingStates QWaylandWindow::toplevelWindowTilingStates() const
+{
+    return mLastReportedToplevelWindowTilingStates;
+}
+
+void QWaylandWindow::handleToplevelWindowTilingStatesChanged(ToplevelWindowTilingStates states)
+{
+    mLastReportedToplevelWindowTilingStates = states;
+}
+
+Qt::WindowStates QWaylandWindow::windowStates() const
+{
+    return mLastReportedWindowStates;
+}
+
 void QWaylandWindow::handleWindowStatesChanged(Qt::WindowStates states)
 {
     createDecoration();
@@ -1198,6 +1375,9 @@ void QWaylandWindow::handleUpdate()
         return;
 
     if (mFrameCallback) {
+        if (!isExposed())
+            return;
+
         wl_callback_destroy(mFrameCallback);
         mFrameCallback = nullptr;
     }
@@ -1277,3 +1457,5 @@ void QWaylandWindow::setOpaqueArea(const QRegion &opaqueArea)
 }
 
 QT_END_NAMESPACE
+
+#include "moc_qwaylandwindow_p.cpp"
